from type_t import T
from typing import Callable, Collection, Iterator


def interact(collection: Collection[T], interaction_func: Callable[[T, Collection[T]], T]) -> Iterator[T]:
    """
    Apply the `interaction_func` to each item in the collection `collection`, 
    where `interaction_func` generates a new item based on the current item and the
    remaining items in the collection. *This is a generator function!*

    Args:
    - `collection: Collection[T]`: The collection of items to modifiy.
    - `interaction_func: Callable[[T, Collection[T]], T]`: The interaction 
      function, which must take an item and a collection of items, returning
      a new (modified) item. The `interact()` function handles excluding the
      current item from the collection in each iteration.

    Yields:
    - `_: Iterator[T]`: The modified items after interaction.

    Notes:
    - This may be useful in things like n-body simulations. 

    Example:
    - Consider a scenario where you have a list of numbers, and you want to apply an
      interaction function that replace each number with the sum of all the others,
      divided by the number itself.

        ```python
        def add_and_divide(item, others):
            return sum(others) / item

        numbers = [1, 2, 3, 4, 5]
        modified_numbers = list(interact(numbers, add_and_divide))
        print(modified_numbers) # Output: [15, 7.5, 5, 3.75, 3]
        ```
    """
    
    yield from (interaction_func(item, [*zip(*filter(lambda i: i[0] != index, enumerate(collection)))[1]]) for index, item in enumerate(collection))

    # alternatively, if you're not as cool...
    # for index, item in enumerate(collection):
    #     r = filter(lambda i: i[0] != index, enumerate(collection)) # filter out the current item from the collection
    #     r = [*zip(*r)[1]]                                          # retrieve the remaining items
    #     r = interaction_func(item, r)                              # apply the interaction
    #     yield r



import itertools      as  it
import more_itertools as mit
import operator       as op
from   type_t         import T
from   typing         import Callable


def fixed_point(base_state: T, evolution_func: Callable[[T], T], compare_func: Callable[[T, T], bool] = op.eq) -> T:
    """
    Generates the fixed point of a function starting from a base state. 

    Args:
    - `evolution_func: Callable`: The function which is repeatedly applied.
    - `base_state: T`: The starting state for the evolution function. 
       Must implement `compare_func`. 
       Defaults to `__eq__`. 

    Returns:
    - `_: T`: The fixed point of the evolved function.

    Notes:
    - May iterate indefinitely. This is considered user error.
    """

    return next(filter(lambda ab: ab[0] == ab[1], mit.windowed(it.islice(it.accumulate(it.repeat(base_state), lambda x, _: evolution_func(x)), 1, None), 2)))[0]

    # alternatively, if you're not as cool...  
    # r = it.accumulate(it.repeat(base_state), lambda x, _: evolution_func(x)) # functional way of calling the `evolution_func` successively on `base_state`
    # r = it.islice(r, 1, None)                                                # chop off the first result because it's the `base_state`, not an iterator
    # r = mit.windowed(r, 2)                                                   # moving window, such that you can use `compare_func` current and previous
    # r = filter(lambda ab: compare_func(ab[0], ab[1]), r)                     # only return windows which pass the `compare_func`
    # r = next(r)[0]                                                           # return the left side of the first such window, i.e. the fixed point
    # return r